/**
 * $Id: unwrapper.c,v 1.9 2005/04/23 01:36:08 blendix Exp $
 *
 * ***** BEGIN GPL/BL DUAL LICENSE BLOCK *****
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version. The Blender
 * Foundation also sells licenses for use in proprietary software under
 * the Blender License.  See http://www.blender.org/BL/ for information
 * about this.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * The Original Code is Copyright (C) 2001-2002 by NaN Holding BV.
 * All rights reserved.
 *
 * The Original Code is: all of this file.
 *
 * Contributor(s): none yet.
 *
 * ***** END GPL/BL DUAL LICENSE BLOCK *****
 */

#include <string.h>
#include <stdlib.h>
#include <math.h>
#include "nl.h"
#include "lscm_unwrap.h"

/* Implementation Least Squares Conformal Maps parameterization, based on
 * chapter 2 of:
 * Bruno Levy, Sylvain Petitjean, Nicolas Ray, Jerome Maillot. Least Squares
 * Conformal Maps for Automatic Texture Atlas Generation. In Siggraph 2002,
 * July 2002.
 */
 
/* Data structure defines */
#define LSCM_SEAM1   1
#define LSCM_SEAM2   2
#define LSCM_INDEXED 4
#define LSCM_PINNED  8

#define MEM_freeN					free
#define MEM_callocN(size,cnt,name)	calloc(size,cnt)
#define MEM_mallocN(size,name)		malloc(size)

#define MAX2(x,y)               ( (x)>(y) ? (x) : (y) )
#define MIN2(x,y)               ( (x)>(y) ? (y) : (x) )
#define INIT_MINMAX(min, max) (min)[0]= (min)[1]= (min)[2]= 1.0e30f; (max)[0]= (max)[1]= (max)[2]= -1.0e30f;
#define INIT_MINMAX2(min, max) (min)[0]= (min)[1]= 1.0e30f; (max)[0]= (max)[1]= -1.0e30f;

#define DO_MINMAX(vec, min, max) \
	if( (min)[0]>(vec)[0] ) (min)[0]= (vec)[0]; \
	if( (min)[1]>(vec)[1] ) (min)[1]= (vec)[1]; \
	if( (min)[2]>(vec)[2] ) (min)[2]= (vec)[2]; \
	if( (max)[0]<(vec)[0] ) (max)[0]= (vec)[0]; \
	if( (max)[1]<(vec)[1] ) (max)[1]= (vec)[1]; \
	if( (max)[2]<(vec)[2] ) (max)[2]= (vec)[2]; 

#define DO_MINMAX2(vec, min, max) \
	if( (min)[0]>(vec)[0] ) (min)[0]= (vec)[0]; \
	if( (min)[1]>(vec)[1] ) (min)[1]= (vec)[1]; \
	if( (max)[0]<(vec)[0] ) (max)[0]= (vec)[0]; \
	if( (max)[1]<(vec)[1] ) (max)[1]= (vec)[1];

enum LSCMFlags_e{
	TF_SEAM1,
	TF_SEAM2,
	TF_SEAM3,
	TF_SEAM4,
	TF_SELECT,
	TF_HIDE,
	TF_PIN1,
	TF_PIN2,
	TF_PIN3,
	TF_PIN4,
	ME_SEAM,
};

/* LscmVert = One UV */
typedef struct LscmVert {
	int v, v1, v2;            /* vertex indices */
	int index;                /* index in solver */
	short tf_index;           /* index in tface (0, 1, 2 or 3) */
	short flag;               /* see above LSCM constants */
	MNMapFace *tf;               /* original tface */
} LscmVert;

/* QuickSort helper function, sort by vertex id */
static int comp_lscmvert(const void *u1, const void *u2)
{
	LscmVert *v1, *v2;
	
	v1= *((LscmVert**)u1);
	v2= *((LscmVert**)u2);

	if (v1->v > v2->v) return 1;
	else if (v1->v < v2->v) return -1;
	return 0;
}




// LOCALS
typedef struct LSCMData_s{ 
	int edgeflag;
	int channel;
	BitArray lockedvertices;
	BitArray *pLockvertices;
	BitArray *pSelfaces;
	MNMap	texMap;
}LSCMData_t;

typedef struct GroupPack_s{
	float	area_xyz;
	float	area_uv;
	float	area_ratio;
	float	min[2];
	float	max[2];
}GroupPack_t;

static LSCMData_t l_data;



/* divide selected faces in groups, based on seams. note that group numbering
starts at 1 */
static int make_seam_groups(MNMesh *me, int **seamgroups)
{
	int a, b, gid, i;
	MNMapFace *tf, *tface;
	MNFace *mf, *mface;
	int *gf, *gface, *groups;
	int doit, mark;
	BitArray	usededges;

	usededges.SetSize(me->ENum());

	if(!me ) return 0;

	groups= (int*)MEM_callocN(sizeof(int),me->FNum(), "SeamGroups");
	
	mface= me->f;
	tface= l_data.texMap.f;
	gface= groups;
	gid= 0;
	for(b=me->FNum(),i=0; b>0; b--, mface++, tface++, gface++,i++) {
		// only process selected and those who are not already part of a group
		if(*gface!=0 || (l_data.pSelfaces && !l_data.pSelfaces->operator [](i))) continue;

		// for every new group clear table
		if(gid != 0)
			usededges.ClearAll();

		gid++;
		*gface= gid;
		mark= 0;
		doit= 1;


		while(doit) {
			doit= 0;

			/* select connected: fill array with those of same groupid */
			tf= tface;
			mf= mface;
			gf= gface;
			a= b;
			while(a--) {
				if(*gf==gid) {
					for(int n = 0; n < mf->deg; n++){
						usededges.Set(mf->edg[n],1); 
					}
				}
				tf++; mf++; gf++;
			}

			/* select the faces using array
			* consider faces connected when they share one non-seam edge */
			tf= tface;
			mf= mface;
			gf= gface;
			a= b;
			int findex = i;
			while(a--) {
				if(*gf==0 && (!l_data.pSelfaces || (l_data.pSelfaces && l_data.pSelfaces->operator [](findex)))) {
					mark= 0;

					for(int n = 0; n < mf->deg; n++){
						if(!me->e[mf->edg[n]].GetFlag(l_data.edgeflag) &&
							usededges[mf->edg[n]])
							mark = 1;
					}

					if(mark) {
						*gf= gid;
						doit= 1;
					}
				}
				tf++; mf++; gf++; findex++;
			}
		}
	}

	*seamgroups= groups;

	return gid;
}

static void lscm_rotate_vert(int a, LscmVert **sortvert, int v2, int index)
{
	LscmVert **sv, *v;
	int found, b;

	/* starting from edge sortvert->v,v2, rotate around vertex and set
	* index until a seam or an already marked tri is encountered */
	found = 1;

	while(found) {
		found= 0;
		sv=sortvert;

		for(b=a; b>0 && ((*sv)->v == (*sortvert)->v) && !found; b--, sv++) {
			v= *sv;

			if(v->flag & LSCM_INDEXED);
			else if(v->v1 == v2) {
				v2= v->v2;

				if(v->flag & LSCM_SEAM1) break;

				v->index= index;
				v->flag |= LSCM_INDEXED;
				found= 1;
				break;
			}
			else if(v->v2==v2) {
				v2= v->v1;

				if(v->flag & LSCM_SEAM2) break;

				v->index= index;
				v->flag |= LSCM_INDEXED;
				found= 1;
				break;
			}
		}
	}
}

static int lscm_vertex_set_index(int a, LscmVert **sortvert, int totindex)
{
	LscmVert **sv, *v;
	int index, b;

	/* rotate over 'wheel' of faces around vertex, incrementing the index
	everytime we meet a seam, or no next connected face is found.
	repeat this until we have and id for all verts.
	if mesh is non-manifold, non-manifold edges will be cut randomly */

	index= totindex;
	sv= sortvert;

	for(b=a; b>0 && ((*sv)->v == (*sortvert)->v); b--, sv++) {
		v= *sv;

		if(v->flag & LSCM_INDEXED) continue;

		v->index= index;
		v->flag |= LSCM_INDEXED;

		lscm_rotate_vert(b, sv, v->v1, index);
		lscm_rotate_vert(b, sv, v->v2, index);

		index++;
	}

	return index;
}

static int lscm_set_indices(LscmVert **sortvert, int totvert)
{
	LscmVert *v, **sv;
	int a, lastvert, totindex;

	totindex= 0;
	lastvert= -1;
	sv= sortvert;

	for(a=totvert; a>0; a--, sv++) {
		v= *sv;
		if(v->v != lastvert) {
			totindex= lscm_vertex_set_index(a, sv, totindex);
			lastvert= v->v;
		}
	}

	return totindex;
}

static void lscm_normalize(Point3 &co, Point3 &center, float radius)
{
	/* normalize relative to complete surface */
	co = co-center;
	co *= (float)1.0/radius;
}

static void lscm_add_triangle(Point3 &v1, Point3 &v2, Point3 &v3, int vid1, int vid2, int vid3, Point3 &center, float radius)
{
	Point3 x, y, z, sub;
	float z1[2], z2[2];
	int id0[2], id1[2], id2[2];

	/* project 3d triangle
	* edge length is lost, as this algorithm is angle based */
	lscm_normalize(v1, center, radius);
	lscm_normalize(v2, center, radius);
	lscm_normalize(v3, center, radius);

	x = v2 -v1;
	x.Unify();


	sub = v3-v1;
	z = CrossProd(x,sub);
	z.Unify();

	y = CrossProd(z,x);

	/* reduce to two 2d vectors */
	sub = v2 - v1;
	z1[0]= sub.Length();
	z1[1]= 0;

	sub = v3 - v1;
	z2[0]= DotProd(sub,x);
	z2[1]= DotProd(sub, y);

	/* split id's up for u and v
	id = u, id + 1 = v */
	id0[0]= 2*vid1;
	id0[1]= 2*vid1 + 1;
	id1[0]= 2*vid2;
	id1[1]= 2*vid2 + 1;
	id2[0]= 2*vid3;
	id2[1]= 2*vid3 + 1;

	/* The LSCM Equation:
	* ------------------
	* (u,v) are the uv coords we are looking for -> complex number u + i*v
	* (x,y) are the above calculated local coords -> complex number x + i*y
	* Uk = uk + i*vk
	* Zk = xk + i*yk (= zk[0] + i*zk[1] in the code)
	* 
	* That makes the equation:
	* (Z1 - Z0)(U2 - U0) = (Z2 - Z0)(U1 - U0)
	*
	* x0, y0 and y1 were made zero by projecting the triangle:
	* (x1 + i*y1)(u2 + i*v2 - u0 - i*v0) = (x2 + i*y2)(u1 + i*v1 - u0 - i*v0)
	*
	* this gives the following coefficients:
	* u0 * ((-x1 + x2) + i*(y2))
	* v0 * ((-y2) + i*(-x1 + x2))
	* u1 * ((-x2) + i*(-y2))
	* v1 * ((y2) + i*(-x2))
	* u2 * (x1)
	* v2 * (i*(x1))
	*/

	/* real part */
	nlBegin(NL_ROW);
	nlCoefficient(id0[0], -z1[0] + z2[0]);
	nlCoefficient(id0[1], -z2[1]        );
	nlCoefficient(id1[0], -z2[0]        );
	nlCoefficient(id1[1],  z2[1]        );
	nlCoefficient(id2[0],  z1[0]        );
	nlEnd(NL_ROW);

	/* imaginary  part */
	nlBegin(NL_ROW);
	nlCoefficient(id0[0],  z2[1]        );
	nlCoefficient(id0[1], -z1[0] + z2[0]);
	nlCoefficient(id1[0], -z2[1]        );
	nlCoefficient(id1[1], -z2[0]        );
	nlCoefficient(id2[1],  z1[0]        );
	nlEnd(NL_ROW);
}

static float lscm_angle_cos(Point3 &v1, Point3 &v2, Point3 &v3)
{
	Point3 vec1, vec2;

	vec1 = v2 - v1;
	vec2 = v3 - v1;
	vec1.Unify();
	vec2.Unify();

	return DotProd(vec1,vec2);
}

static int lscm_build_vertex_data(MNMesh *me, int *groups, int gid, LscmVert **lscm_vertices, LscmVert ***sort_vertices)
{
	MNVert *mv;
	MNFace *mf;
	MNMapFace *tf;
	int *gf, totvert, a;
	int e;
	LscmVert *lscmvert, **sortvert;
	LscmVert *v1, *v2, *v3, **sv1, **sv2, **sv3;
	float a1, a2;
	int n;
	int uvoffset;

	/* determine size for malloc */
	totvert= 0;
	mv = me->v;
	mf= me->f;
	gf= groups;
	a1 = a2 = 0;

	for(a=me->FNum(); a>0; a--) {
		if(*gf==gid) {

			// unique vertices
			totvert += mf->TriNum()*3;
		}
		mf++; gf++;
	}

	/* a list per face vertices */
	lscmvert= (LscmVert*)MEM_callocN(sizeof(LscmVert),totvert,"LscmVerts");
	/* the above list sorted by vertex id */
	sortvert= (LscmVert**)MEM_callocN(sizeof(LscmVert*),totvert, "LscmVSort");

	/* actually build the list (including virtual triangulation) */
	mf= me->f;
	tf= l_data.texMap.f;
	gf= groups;

	v1= lscmvert;
	v2= lscmvert + 1;
	v3= lscmvert + 2;

	sv1= sortvert;
	sv2= sortvert + 1;
	sv3= sortvert + 2;

	// debug
	BitArray seamedges;
	me->getEdgesByFlag(seamedges,l_data.edgeflag,0);

	/* warning: ugly code :) */
	uvoffset = 0;
	for(a=me->FNum(); a>0; a--) {
		if(*gf==gid) {
			Tab<int> triVerts;
			mf->GetTriangles(triVerts);
			for (n = 0; n < mf->TriNum()*3; n+=3)
			{
				v1->v= mf->vtx[triVerts[n]];
				v2->v= mf->vtx[triVerts[n+1]];
				v3->v= mf->vtx[triVerts[n+2]];

				v1->tf_index= tf->tv[triVerts[n]];
				v2->tf_index= tf->tv[triVerts[n+1]];
				v3->tf_index= tf->tv[triVerts[n+2]];

				v1->flag= v2->flag= v3->flag= 0;

				v1->v1= v2->v;
				v1->v2= v3->v;

				v2->v1= v1->v;
				v2->v2= v3->v;

				v3->v1= v1->v;
				v3->v2= v2->v;

				v1->tf= v2->tf= v3->tf= tf;

				*sv1= v1;
				*sv2= v2;
				*sv3= v3;

				e=me->FindEdgeFromVertToVert(v1->v,v2->v);
				if(e >= 0 && seamedges[e]) {
					v1->flag |= LSCM_SEAM1;
					v2->flag |= LSCM_SEAM1;
				}

				e=me->FindEdgeFromVertToVert(v2->v,v3->v);
				if(e >= 0 && seamedges[e]) {
					v2->flag |= LSCM_SEAM2;
					v3->flag |= LSCM_SEAM2;
				}

				e=me->FindEdgeFromVertToVert(v1->v,v3->v);
				if(e >= 0  && seamedges[e]) {
					v1->flag |= LSCM_SEAM2;
					v3->flag |= LSCM_SEAM1;
				}

				v1 += 3; v2 += 3; v3 += 3;
				sv1 += 3; sv2 += 3; sv3 += 3;
			}
		}
		tf++; mf++; gf++;
	}


	/* sort by vertex id */
	qsort(sortvert, totvert, sizeof(LscmVert*), comp_lscmvert);

	*lscm_vertices= lscmvert;
	*sort_vertices= sortvert;
	return totvert;
}

static void lscm_min_max_cent_rad(MNMesh *me, LscmVert **sortvert, int totvert, float *min, float *max, Point3 &center, float *radius)
{
	MNVert *mv= me->v;
	LscmVert *v, **sv;
	int a, lastvert, vertcount;
	float *co;
	Point3 sub;

	/* find min, max and center */
	center.Set(0,0,0);
	INIT_MINMAX(min, max);

	vertcount= 0;
	lastvert= -1;
	sv= sortvert;

	for(a=totvert; a>0; a--, sv++) {
		v= *sv;
		if(v->v != lastvert) {
			co= (float*)((MNVert *)(mv+v->v))->p;

			center += (mv+v->v)->p;
			DO_MINMAX(co, min, max);

			vertcount++;
			lastvert= v->v;
		}
	}

	center *= (float)1.0/(float)vertcount;

	/* find radius */
	sub = center - sub.Set(max[0],max[1],max[2]);
	*radius= sub.Length();

	if(*radius < 1e-20)
		*radius= 1.0;
}

static void lscm_projection_axes(float *min, float *max,Point3 &p1, Point3 &p2)
{
	float dx, dy, dz;

	dx= max[0] - min[0];
	dy= max[1] - min[1];
	dz= max[2] - min[2];

	p1.Set(0,0,0);
	p2.Set(0,0,0);

	if(dx < dy && dx < dz) {
		if(dy > dz) p1.y= p2.z= 1.0;   /* y, z */
		else p1.x= p2.y= 1.0;          /* z, y */
	}
	else if(dy < dx && dy < dz) {
		if(dx > dz) p1.x= p2.z= 1.0;   /* x, z */
		else p1.z= p2.x= 1.0;          /* z, x */
	}
	else {
		if(dx > dy) p1.x= p2.y= 1.0;   /* x, y */
		else p1.y= p2.x= 1.0;          /* y, x */
	}
}

static void lscm_set_initial_solution(MNMesh *me, LscmVert **sortvert, int totvert, Point3 &p1, Point3 &p2, int *vertex_min, int *vertex_max)
{
	float umin, umax;
	Point3 co;
	int vmin, vmax, a;
	LscmVert **sv, *v;
	MNVert *mv= me->v;
	MNMap *tmap = &l_data.texMap;

	umin= 1.0e30f;
	umax= -1.0e30f;

	vmin= 0;
	vmax= 2;

	sv= sortvert;

	for(a=totvert; a>0; a--, sv++) {
		v= *sv;
		co= me->V(v->v)->p;
		Point3 &uv = tmap->v[v->tf_index];

		uv.x= DotProd(co, p1);
		uv.y= DotProd(co, p2);
		uv.z = 0.0f;

		if(uv.x < umin) {
			vmin= v->index;
			umin= uv.x;
		}
		if(uv.x > umax) {
			vmax= v->index;
			umax= uv.x;
		}

		nlSetVariable(2*v->index, uv.x);
		nlSetVariable(2*v->index + 1, uv.y);
	}

	*vertex_min= vmin;
	*vertex_max= vmax;
}

static void lscm_set_pinned_solution(MNMesh *me, LscmVert **sortvert, int totvert, int *pinned)
{
	float min[2], max[2], *uv;
	int a, pin;
	LscmVert **sv, *v;
	MNVert *mv= me->v;

	INIT_MINMAX2(min, max);
	*pinned= 0;

	sv= sortvert;

	for(a=totvert; a>0; a--, sv++) {
		v= *sv;
		uv= (float*)l_data.texMap.v[v->tf_index];

		pin = l_data.lockedvertices[v->tf_index];

		nlSetVariable(2*v->index, uv[0]);
		nlSetVariable(2*v->index + 1, uv[1]);

		if(pin){
			DO_MINMAX2(uv, min, max);

			*pinned += 1;

			nlLockVariable(2*v->index);
			nlLockVariable(2*v->index + 1);
		}
	}

	if (*pinned){
		/* abuse umax vmax for caculating euclidian distance */
		max[0] -= min[0];
		max[1] -= min[1];

		/* check for degenerated pinning box */
		if (((max[0]*max[0])+(max[1]*max[1])) < 1e-10)
			*pinned = -1;
	}
}


static void lscm_build_matrix(MNMesh *me, LscmVert *lscmvert, int *groups, int gid, Point3 &center, float radius)
{
	MNVert *mv = me->v;
	MNFace *mf;
	int *gf, a, id1, id2, id3;
	LscmVert *v;
	Point3 co1, co2, co3;

	nlBegin(NL_MATRIX);

	mf= me->f;
	gf= groups;
	v= lscmvert;

	for(a=me->FNum(); a>0; a--) {
		if(*gf==gid) {
			for(int n = 0; n < mf->TriNum(); n++){
				co1 = (mv+v->v)->p;
				id1= v->index; v++;
				co2 = (mv+v->v)->p;
				id2= v->index; v++;
				co3 = (mv+v->v)->p;
				id3= v->index; v++;

				lscm_add_triangle(co1, co2, co3, id1, id2, id3, center, radius);
			}
		}
		mf++; gf++;
	}

	nlEnd(NL_MATRIX);
}

static void lscm_load_solution(MNMesh *me, LscmVert *lscmvert, int *groups, int gid)
{
	MNFace *mf;
	MNMap *tmap = &l_data.texMap;
	//MNMapFace *tf = me->MF(l_data.channel,0);
	int *gf, a;
	LscmVert *v;

	mf= me->f;
	gf= groups;
	v= lscmvert;

	for(a=me->FNum(); a>0; a--) {
		if(*gf==gid) {

			/* index= u, index + 1= v */
			for(int n = 0; n < mf->TriNum()*3; n++){
				Point3 &uv = tmap->v[v->tf_index];

				uv[0]= (float)nlGetVariable(2*v->index);
				uv[1]= (float)nlGetVariable(2*v->index + 1);
				uv[2]= 0.0f;

				v++;
			}
			
		}
		mf++; gf++; 
	}
}

static int unwrap_lscm_face_group(MNMesh *me, int *groups, int gid)
{
	LscmVert *lscmvert, **sortvert;
	int totindex, totvert, vmin, vmax,pinned;
	float min[3], max[3], radius;
	Point3 p1,p2;
	Point3 center;

	/* build the data structures */
	totvert= lscm_build_vertex_data(me, groups, gid, &lscmvert, &sortvert);

	/* calculate min, max, center and radius */
	lscm_min_max_cent_rad(me, sortvert, totvert, min, max, center, &radius);

	/* index distinct vertices */
	totindex= lscm_set_indices(sortvert, totvert);

	/* create solver */
	nlNewContext();
	nlSolverParameteri(NL_NB_VARIABLES, 2*totindex);
	nlSolverParameteri(NL_LEAST_SQUARES, NL_TRUE);

	nlBegin(NL_SYSTEM);

	///* find axes for projecting initial solutions on */
	lscm_projection_axes(min, max, p1, p2);
	///* see if pinned data is avail and set on fly */
	lscm_set_pinned_solution(me, sortvert, totvert, &pinned);

	if(pinned < 0); /* really small pinned uv's: won't see difference anyway */
	else { 
		/* auto pinning */
		if(pinned < 2) 
		{
			/* set initial solution and locate two extrema vertices to pin */
			lscm_set_initial_solution(me,sortvert,totvert,p1,p2,&vmin,&vmax);

			/* pin 2 uv's */
			nlLockVariable(2*vmin);
			nlLockVariable(2*vmin + 1);
			nlLockVariable(2*vmax);
			nlLockVariable(2*vmax + 1);
		}

		/* add triangles to the solver */
		lscm_build_matrix(me, lscmvert, groups, gid, center, radius);

		nlEnd(NL_SYSTEM);

		/* LSCM solver magic! */
		nlSolve();

		/* load new uv's: will be projected uv's if solving failed  */
		lscm_load_solution(me, lscmvert, groups, gid);
	}

	nlDeleteContext(nlGetCurrent());
	MEM_freeN(lscmvert);
	MEM_freeN(sortvert);
	return (pinned);
}

static void seam_group_bbox(int *groups, int gid, float *min, float *max)
{
	MNMapFace *tf;
	MNMap *tmap = &l_data.texMap;
	int *gf, a;
	float *uv;

	INIT_MINMAX2(min, max);

	tf= l_data.texMap.f;
	gf= groups;

	for(a=tmap->numf; a>0; a--) {
		if((gid!=0 && *gf==gid) || (gid==0 && *gf)) {

			for(int n = 0; n < tf->deg; n++){
				uv= (float*)tmap->v[tf->tv[n]];
				DO_MINMAX2(uv, min, max);
			}
		}
		tf++; gf++;
	}
}

static void seam_group_scale(int *groups, int gid, float scale)
{
	MNMapFace *tf;
	MNMap *tmap = &l_data.texMap;
	int *gf, a;

	tf= l_data.texMap.f;
	gf= groups;

	for(a=tmap->numf; a>0; a--) {
		if((gid!=0 && *gf==gid) || (gid==0 && *gf)) {

			for(int n = 0; n < tf->deg; n++){
				tmap->v[tf->tv[n]] *= scale;
			}
		}
		tf++; gf++;
	}
}

static void seam_group_move(int *groups, int gid, float add[2])
{
	MNMapFace *tf;
	MNMap *tmap = &l_data.texMap;
	int *gf, a;

	tf= l_data.texMap.f;
	gf= groups;

	for(a=tmap->numf; a>0; a--) {
		if((gid!=0 && *gf==gid) || (gid==0 && *gf)) {

			for(int n = 0; n < tf->deg; n++){
				tmap->v[tf->tv[n]] += Point3((float)add[0],(float)add[1],(float)0);
			}
		}
		tf++;gf++;
	}
}

/* put group withing (0,0)->(1,1) boundbox */
static void seam_group_normalize(int *groups, int gid)
{
	float min[2], max[2], sx, sy, scale, add[2];

	seam_group_bbox(groups, gid, min, max);

	sx= (max[0]-min[0]);
	sy= (max[1]-min[1]);

	scale= MAX2(sx, sy);
	scale= (1.0f/scale);

	add[0]= -min[0];
	add[1]= -min[1];

	seam_group_move(groups, gid, add);
	seam_group_scale(groups, gid, scale);
}

/* get scale relative to mesh */
static void seam_group_relative_scale(MNMesh *me, int *groups, int gid,GroupPack_t *pack)
{
	MNVert *mv= me->v;
	MNFace *mf;
	MNMapFace *tf;
	UVVert *uv = l_data.texMap.v;
	Point3	pt1;
	Point3	pt2;
	Point3	pt3;
	int *gf, a;
	float area_xyz, area_uv;

	area_xyz= 0.0;
	area_uv= 0.0;
	mf= me->f;
	tf= l_data.texMap.f;
	gf= groups;

	for(a=me->FNum(); a>0; a--) {
		if(*gf==gid) {
			Tab<int> triVerts;
			mf->GetTriangles(triVerts);
			for (int n = 0; n < mf->TriNum()*3; n+=3)
			{
				pt1 = mv[mf->vtx[triVerts[n]]].p;
				pt2 = mv[mf->vtx[triVerts[n+1]]].p;
				pt3 = mv[mf->vtx[triVerts[n+2]]].p;

				area_xyz += CrossProd((pt2-pt1),(pt3-pt1)).Length() * 0.5f;

				pt1 = uv[tf->tv[triVerts[n]]];
				pt2 = uv[tf->tv[triVerts[n+1]]];
				pt3 = uv[tf->tv[triVerts[n+2]]];

				area_uv += CrossProd((pt2-pt1),(pt3-pt1)).Length() * 0.5f;
			}
		}
		tf++; mf++; gf++;
	}

	pack->area_uv = area_uv;
	pack->area_xyz = area_xyz;
	pack->area_ratio = (area_xyz/area_uv);
}

/* very primitive packing */
static void pack_seam_groups(MNMesh *me, int *groups, int totgroup, float packmax)
{
	float scale, add[2], groupw;
	float dx, dy, packx, packy, min[2], max[2], rowh;
	GroupPack_t *grouppacks;
	GroupPack_t *pack;
	int a;

	grouppacks = (GroupPack_t*)MEM_mallocN(sizeof(GroupPack_t)*totgroup, "SeamGroupScale");


	float totalsize_uv = 0;
	float totalsize_xyz = 0;
	float totalmax = 0;
	pack = grouppacks;
	for(a=0; a<totgroup; a++,pack++) {
		seam_group_relative_scale(me, groups, a+1,pack);
		totalsize_uv += pack->area_uv;
		totalsize_xyz += pack->area_xyz;
		totalmax = totalmax < pack->area_xyz ? pack->area_xyz : totalmax;
	}

	packx= packy= 0.0;
	rowh= 0.0;
	groupw= 1.0f/sqrt((float)totgroup);

	float areascale = 1.0f/sqrtf(totalsize_xyz);

	pack = grouppacks;
	for(a=0; a<totgroup; a++,pack++) {

		/* scale so all groups have the same size relative to the mesh */

		scale = packmax*sqrtf((1.0f/pack->area_uv)/(totalmax/pack->area_xyz));
	
		seam_group_bbox( groups, a+1, min, max);
		dx= (max[0]-min[0])*scale;
		dy= (max[1]-min[1])*scale;

		/* for padding */
		dx += 0.01f;
		dy += 0.01f;

		add[0]= add[1]= 0.0f;

		if(dx > 1.0) {
			add[0]= 0.0f;
			add[1]= packy;

			packy += dy;
			packx= 0.0f;
			rowh= 0.0f;
		}
		else if(dx <= (1.0-packx)) {
			add[0]= packx;
			add[1]= packy;

			packx += dx;
			rowh= MAX2(rowh, dy);
		}
		else {
			packy += rowh;
			packx= dx;
			rowh= dy;

			add[0]= 0.0f;
			add[1]= packy;
		}

		/* for padding */
		add[0] += 0.005f;
		add[1] += 0.005f;

		seam_group_scale( groups, a+1, scale);
		seam_group_move( groups, a+1, add);
	}

	MEM_freeN(grouppacks);
	
/*
	seam_group_normalize( groups, 0);
	seam_group_scale( groups, 0, 0.98f);
	add[0]= add[1]= 0.01f;
	seam_group_move( groups, 0, add);
	*/
}

static void pack_seam_groups_nonscale(int *groups, int totgroup)
{
	int width = sqrt((float)totgroup) + 1;
	float scale = 1.0f/width;
	float add[2];

	add[0] = 0.0f;
	add[1] = 0.0f;
	for (int a=0; a < totgroup; a++){
		seam_group_scale(groups,a+1,scale);
		seam_group_move(groups,a+1,add);
		add[0] += 1.0f*scale + 0.005f;
		if (a && (a%width)==0){
			add[0] = 0.0f ;
			add[1] += 1.0f*scale + 0.005f;
		}
	}
}

// this functions prepares map verts/faces to create unique vertices per face
// also remembers old values
// it sets internal map
void mesh_prepMap(MNMesh *me, const MNMap *tmap)
{
	
	int totalvert = 0;

	l_data.texMap.ClearAndFree();
	l_data.texMap = *tmap;
	if (l_data.texMap.f == tmap->f)
		l_data.texMap.NewAndCopy();


	// count original tverts
	MNMapFace *tf = tmap->f;
	for (int i = 0; i < tmap->numf; i++,tf++){
		totalvert += tf->deg;
	}

	l_data.lockedvertices.SetSize(totalvert);
	l_data.lockedvertices.ClearAll();

	if (l_data.pSelfaces || l_data.pLockvertices){
		UVVert *pOldpoints = tmap->v;

		l_data.texMap.setNumVerts(totalvert);

		// reindex faces, unique per face

		totalvert = 0;
		tf = l_data.texMap.f;
		for (int i = 0; i < l_data.texMap.numf; i++,tf++){
			for (int n = 0; n < tf->deg; n++){
				// copy oldpoint to new index position
				l_data.texMap.v[totalvert] = pOldpoints[tf->tv[n]];
				if (l_data.pLockvertices && l_data.pLockvertices->operator [](tf->tv[n]))
					l_data.lockedvertices.Set(totalvert);
				tf->tv[n] = totalvert++;
			}
		}

	}
	else{
		l_data.texMap.setNumVerts(totalvert);

		// reindex faces, unique per face

		totalvert = 0;
		tf = l_data.texMap.f;
		for (int i = 0; i < l_data.texMap.numf ; i++,tf++){
			for (int n = 0; n < tf->deg ; n++){
				tf->tv[n] = totalvert++;
			}
		}
	}
}


static void weldSelected(MNMesh *me, MNMap *tmap,float weldThreshold, int *groups, int totgroups)
{
	
	int *gid;
	MNMapFace *tf;
	float sweldThreshold = weldThreshold * weldThreshold;
	BitArray deadverts;
	BitArray mergedverts;
	BitArray groupverts;

	deadverts.SetSize(tmap->VNum());
	mergedverts.SetSize(tmap->VNum());
	groupverts.SetSize(tmap->VNum());

	// we weld based on groups
	for (int grp = totgroups; grp >= 0; grp--){
		tf = tmap->f;
		gid = groups;

		// find out all vertices that belong to this group
		for (int f = 0; f < tmap->numf; f++,tf++,gid++){
			if (*gid == grp){
				for (int n = 0; n < tf->deg; n++)
					groupverts.Set(tf->tv[n]);
			}
		}

		// range check all verts of this group
		for (int m=0; m < tmap->VNum(); m++) 
		{
			if (deadverts[m] || !(groupverts[m]))
				continue;

			Point3 p(0.0f,0.0f,0.0f);
			Point3 op(0.0f,0.0f,0.0f);
			p = tmap->V(m);
			op = p;
			int ct = 0;
			int index = -1;
			int i;
			for (i=m+1; i < tmap->VNum(); i++) 
			{
				if ( !(deadverts[i]) && (groupverts[i]))
				{
					Point3 np;

					np= tmap->V(i);
					if (LengthSquared(np-op) < sweldThreshold)
					{
						p+= np;
						ct++;
						if (index == -1)
							index = m;
						deadverts.Set(i);
					}
				}
			}

			if (!((index == -1) || (ct == 0)))
			{
				ct++;
				p = p /(float)ct;
				mergedverts.Set(index);

				tmap->v[index] = p;

				MNMapFace *tf = tmap->f;
				for (i=0; i< tmap->FNum(); i++,tf++) 
				{
					for (int j=0; j < tf->deg; j++) 
					{
						if ( deadverts[tf->tv[j]])
							tf->tv[j] = index;
					}
				}
			}
		}

		groupverts.ClearAll();
	}




	tmap->CollapseDeadVerts(me->f);

}

static void feed_unwrap(IUnwrapMod *unwrap, TimeValue timev)
{
	int i;
	MNMapFace *tf = l_data.texMap.f;
	int numoriginalverts = unwrap->fnNumberVertices();

	// fill vertices
	UVVert *uvpt = l_data.texMap.v;
	for (i = 1; i < numoriginalverts+1; i++,uvpt++){
		unwrap->fnSetVertexPosition(timev,i,*uvpt);
	}

	// fill faces
	int offset = 0;
	for (i = 1; i < l_data.texMap.numf+1; tf++,i++){
		int *vindex = tf->tv;
		for (int n = 1; n < tf->deg+1; n++,vindex++){
			// create new points if needed
			if (*vindex >= numoriginalverts)
				unwrap->fnAddPoint(l_data.texMap.v[*vindex],i+offset,n+offset,false);
			else
				unwrap->fnSetFaceVertexIndex(i+offset,n+offset,(*vindex + 1)+offset);
		}
	}

}

void unwrap_lscm(MNMesh &me,IUnwrapMod *newunwrap, int channel, int edgeflag, BitArray *selfaces, BitArray *lockvertices, float threshold, TimeValue timev,float packmax)
{

	int dopack = 1;
	int res;
	int totgroup, *groups=NULL, a;

	l_data.channel = channel;
	l_data.edgeflag = edgeflag;

	l_data.pSelfaces = selfaces;
	l_data.pLockvertices = lockvertices;


	// in case no mapchannel exists, create one
	// this changes baseobject, which is a bit dirty
	if (!me.M(channel)){
		if (!me.numm)
			me.SetMapNum(channel+1);
		me.InitMap(channel);
	}

	// copies and sets up internal mapping data
	mesh_prepMap(&me,me.M(channel));

	totgroup= make_seam_groups(&me, &groups);

	if(totgroup==0) return ;

	for(a=totgroup; a>0; a--) {
		res= unwrap_lscm_face_group(&me, groups, a);
		
		if((res < 3) && (res > -1)) {
			seam_group_normalize(groups, a);
		}
		else {
			dopack = 0;
		}
		
	}

	//if (dopack) pack_seam_groups_nonscale(groups,totgroup);
	if (dopack) pack_seam_groups(&me, groups, totgroup,packmax);

	weldSelected(&me,&l_data.texMap,threshold,groups,totgroup);

	free(groups);

	
	feed_unwrap(newunwrap,timev);

	return ;
}
